/*
 * Created 2018 by Max Fahl, PIXILAB Technologies AB.
 */

import { PropertyAccessor, Script, ScriptEnv } from 'system_lib/Script';
import { resource } from "../system_lib/Metadata";
import { SimpleFile } from "../system/SimpleFile";
import { SimpleImage } from "../system/SimpleImage";

// noinspection JSUnusedGlobalSymbols
export class Photo extends Script {

//Path of the block to output photo to.
	private readonly kBlockPath = '/public/block/VisitorsPhoto/Photo';

	private mPublishedTask: PropertyAccessor<boolean>;

	public constructor(env: ScriptEnv) {
		super(env);
		this.mPublishedTask = this.getProperty('Realm.VisitorsPhoto.group.Photo.OnPublish.running', () => {});
		console.log('Photo started.');
	}

	@resource()
	public finish(data: { tempPath: string }): any {
		const { tempPath } = data;
		const tempSrc = "/temp/" + tempPath;

		return new Promise((resolve) => {
			SimpleImage.info(tempSrc).then(info => {
				let cropWidth, cropHeight, cropTop, cropLeft;

				// If output block is landscape 1920x1080
			/*	 if (info.height / info.width >= 0.5625) {
				 	cropWidth = info.width;
				 	cropHeight = info.width * 0.5625;
				cropTop = (info.height - cropHeight) / 2;
			 	cropLeft = 0;
			 } else {
				cropHeight = info.height
			 	cropWidth = info.height * 1.777777778;
			 	cropLeft = (info.width - cropWidth) / 2;
		 	cropTop = 0;
				 }
*/

				// If output block is portrait 1080x1920
			if (info.height / info.width >= 1.777777778) {
					cropWidth = info.width;
					cropHeight = info.width * 1.777777778;
					cropTop = (info.height - cropHeight) / 2;
					cropLeft = 0;
				} else {
					cropHeight = info.height;
					cropWidth = info.height * 0.5625;
					cropLeft = (info.width - cropWidth) / 2;
					cropTop = 0;
				}


				SimpleImage.derive(
					tempSrc,
					`${ this.kBlockPath }/media/image.jpg`,
					cropLeft,
					cropTop,
					cropWidth,
					cropHeight,
					1080,
					1920
				).then(() => {
					SimpleFile.read(`${ this.kBlockPath }/Spec.pixi`).then(
						contents => {
							let data = JSON.parse(contents);
							data['root']['blocks'][0]['mMedia']['mFileTime'] = +(new Date());
							SimpleFile.write(
								`${ this.kBlockPath }/Spec.pixi`,
								JSON.stringify(data)
							).then(() => {
								this.mPublishedTask.value = true;
								resolve();
							});
						}
					);
				});
			});
		});
	}
}
